from datetime import timedelta as TimeDelta
from typing import Any, Optional, Protocol

from flask import Flask

from flask_session._utils import total_seconds
from flask_session.base import ServerSideSession, ServerSideSessionInterface
from flask_session.defaults import Defaults

try:
    import waptmemcache
except ImportError:
    from waptlicences import waptmemcache

class WaptMemcacheClientProtocol(Protocol):
    def get(self, key: str) -> Optional[Any]: ...
    def set(self, key: str, value: Any, timeout: int) -> bool: ...
    def delete(self, key: str) -> bool: ...


class WaptMemcacheSession(ServerSideSession):
    pass


class WaptMemcacheSessionInterface(ServerSideSessionInterface):
    """A Session interface that uses memcached as session storage. (`pylibmc`, `libmc`, `python-memcached` or `pymemcache` required)

    :param client: A ``memcache.Client`` instance.
    :param key_prefix: A prefix that is added to all storage keys.
    :param use_signer: Whether to sign the session id cookie or not.
    :param permanent: Whether to use permanent session or not.
    :param sid_length: The length of the generated session id in bytes.
    :param serialization_format: The serialization format to use for the session data.

    """

    serializer = ServerSideSessionInterface.serializer
    session_class = WaptMemcacheSession
    ttl = True

    def __init__(
        self,
        app: Flask,
        port: str = '11212',
        key_prefix: str = Defaults.SESSION_KEY_PREFIX,
        use_signer: bool = Defaults.SESSION_USE_SIGNER,
        permanent: bool = Defaults.SESSION_PERMANENT,
        sid_length: int = Defaults.SESSION_ID_LENGTH,
        serialization_format: str = Defaults.SESSION_SERIALIZATION_FORMAT,
    ):
        self.client = waptmemcache.WaptMemcacheClient(port=str(port))
        super().__init__(
            app, key_prefix, use_signer, permanent, sid_length, serialization_format
        )

    def _retrieve_session_data(self, store_id: str) -> Optional[dict]:
        # Get the saved session (item) from the database
        serialized_session_data = self.client.get(store_id)
        if serialized_session_data:
            return self.serializer.decode(serialized_session_data)
        return None

    def _delete_session(self, store_id: str) -> None:
        self.client.delete(store_id)

    def _upsert_session(
        self, session_lifetime: TimeDelta, session: ServerSideSession, store_id: str
    ) -> None:
        storage_time_to_live = total_seconds(session_lifetime)

        # Serialize the session data
        serialized_session_data = self.serializer.encode(session)

        # Update existing or create new session in the database
        self.client.set(
            store_id,
            serialized_session_data,
            storage_time_to_live,
        )
